from collections import defaultdict
from typing import Any, Callable, List, Tuple

from .models import CRG, CRGData, Event, NaiveGroups, Uid


def crg(
    data: List[Event],
    uids: List[Uid],
    glue: Uid,
    contradiction_solver: Callable[[CRG, Any], List[Event]] = \
        lambda state, uid_val, glue: []
) -> CRGData:
    """
    performs cross reference matching of data by uids
    grouping by grouping marker if present in matched group

    :param data: events for grouping
    :param uids: list of uids
    :param glue: uid for grouping
    :param contradiction_solver: function to solve in level contradictions
    :return: grouping result by glue uid
    """

    # initialize mutable state of
    # the upcoming cross-reference process
    state: CRG = CRG(
        get_naive_groups(data, uids),
        sorted(uids, key = lambda u: u.priority),
    )

    return group_recursively(state, glue, contradiction_solver)


def get_naive_groups(
    events: List[Event],
    uids: List[Uid],
) -> NaiveGroups:
    """
    simple grouping of data by uids;
    based on presence of equal uid values

    :param events: events for naive grouping
    :param uids: list of uids
    :return: grouping result by uids
    """

    groups = defaultdict(lambda: defaultdict(list))

    for uid in uids:
        for event in events:
            val = uid.get_value(event)
            if val:
                groups[uid][val] += [event]

    return groups


def group_recursively(
    state: CRG,
    glue: Uid,
    contradiction_solver: Callable[[CRG, Any, Uid], List[Event]],
) -> CRGData:
    """
    performs procedures recursively grouping filtering and
    solving contradictions filling the result

    :param state: current state
    :param glue: grouping uid
    :return: result mutated from current step
    """

    if state.current_group:
        group_used_ids = []

        for uid_val, uid_val_data in state.current_group.items():
            used_ids, group, glue_val = get_level_group(
                uid_val_data,
                glue,
                state,
                uid_val,
                contradiction_solver,
            )

            if group:
                state.append_group(
                    glue_val,
                    state.current_uid,
                    uid_val,
                    group,
                )
                group_used_ids += used_ids

        # filter matched data from
        # lower priority levels
        state.filter_data(group_used_ids)

        # mutate state to next level uid
        state.mutate()

        return group_recursively(
            state,
            glue,
            contradiction_solver,
        )

    return state.current_result


def get_glue_val(
    group: List[Event],
    glue: Uid
) -> Any:
    return glue.get_value(next(filter(
        lambda e: e.is_glued(glue),
        group
    )))


def get_level_group(
    group: List[Event],
    glue: Uid,
    state: CRG,
    uid_val: Any,
    contradiction_solver: Callable[[CRG, Any, Uid], List[Event]],
) -> Tuple[List[int], List[Event], Any]:
    used_ids = []
    level_group = []
    glue_val = None

    if len(group) > 1:
        group_entropy = get_group_entropy(group, glue)

        # solve contradictions
        if group_entropy > 0:
            group = contradiction_solver(state, uid_val, glue)
            group_entropy = get_group_entropy(group, glue)

        # zero entropy = zero confusion
        if not group_entropy:
            for event in group:

                if not event.is_glued(glue):
                    level_group.append(event)
                    used_ids.append(event.id)

                else:
                    glue_val = glue.get_value(event)

    return used_ids, level_group, glue_val


def get_group_entropy(group: List[Event], glue: Uid) -> int:
    entropy_container = set()

    for event in group:
        glued = event.is_glued(glue)

        if glued:
            entropy_container.update([glue.get_value(event)])

    return len(entropy_container) - 1



from collections import defaultdict
from typing import Any, Callable, Dict, Iterable, List, Optional

class Event:
    def __init__(
        self,
        name: str,
        value: Dict[Any, Any],
        id: Optional[int] = None,
    ):
        # use provided id or
        # or create unique per event contents
        self.id = id
        if not self.id:
            self.id = frozenset(value.items()).__hash__()

        self.name = name
        self._value = value

    def is_glued(self, glue: 'Uid') -> bool:
        return glue.name in self.value

    @property
    def value(self):
        return self._value

    def update_value(self, value: Dict[Any, Any]):
        self._value.update(value)
        self.id = frozenset(self._value.items()).__hash__()

class Uid:
    """
    unique identifier of event
    """

    def __init__(
        self,
        priority: int,
        name: Any,
        value_getter: Optional[Callable[[Event], Any]] = None,
    ):
        self.name = name
        self.priority = priority

        # if no value getter provided, use default one;
        # default value getter gets uid value
        # from corresponding field in event
        self._value_getter = value_getter
        if not self._value_getter:
            self._value_getter = lambda e: e.value[self.name]

    def get_value(self, event: Event) -> Any:
        return self._value_getter(event)

    def __hash__(self) -> int:
        return self.name.__hash__()

CRGData = Dict[str, Dict[Uid, Dict[Any, List[Event]]]]
NaiveGroups = Dict[Uid, Dict[Any, List[Event]]]
NaiveGroup = Dict[Any, List[Event]]

class CRG:
    """
    state of cross-reference grouping;
    possesses a set of methods for validation and
    filtration during recursive usage
    """

    def __init__(
        self,
        groups: NaiveGroups,
        uids: List[Uid],
    ):
        # global mutable vars
        self._groups = groups
        self._unused_uids = iter(sorted(uids, key = lambda u: u.priority))
        self._incremental_result: CRGData = defaultdict(
            lambda: defaultdict(
                lambda: defaultdict(list)
            )
        )

        # level vars
        self._current_uid: Uid = next(self._unused_uids)
        self._current_level: int = self.current_uid.priority
        self._previous_uids: List[Uid] = []
        self._times_mutated = 0

    @property
    def current_result(self):
        return self._incremental_result

    @property
    def current_uid(self) -> Uid:
        return self._current_uid

    @property
    def current_level(self) -> int:
        return self._current_level

    @property
    def current_group(self) -> Optional[NaiveGroup]:
        group = []

        if self._current_uid in self.groups:
            group = self._groups[self._current_uid]

        return group

    @property
    def previous_uids(self) -> Iterable[Uid]:
        return iter(sorted(
            self._previous_uids,
            key = lambda u: u.priority,
            reverse = True,
        ))

    @property
    def times_mutated(self) -> int:
        return self._times_mutated

    @property
    def groups(self) -> NaiveGroups:
        return self._groups

    def filter_data(
        self,
        used_ids: List[int]
    ):
        filtered_groups = defaultdict(lambda: defaultdict(list))

        for mapped_uid, mapped_uid_data in self._groups.items():

            for mapped_uid_val, mapped_uid_val_data in mapped_uid_data.items():
                filtered_group = list(filter(
                    lambda x: not x.id in used_ids,
                    mapped_uid_val_data
                ))

                if filtered_group:
                    filtered_groups[mapped_uid][mapped_uid_val] = \
                        filtered_group

        self._groups = filtered_groups

    def append_group(
        self,
        glue_uid_val: Any,
        uid: Uid,
        uid_val: Any,
        group: List[Event]
    ):
        self._incremental_result[glue_uid_val][uid][uid_val] += group

    def mutate(self):
        self._previous_uids.append(self._current_uid)

        try:
            self._current_uid = next(self._unused_uids)
        except StopIteration:
            self._current_uid = None
            return self

        self._current_level = self._current_uid.priority
        self._times_mutated += 1

        return self


from setuptools import find_packages, setup

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name = "pygrouper",
    version = "1.0",
    author = "bl17zar",
    author_email = "bl17zar@gmail.com",
    description = (
        "simple realisation of cross-reference grouping "
        "algorithm; mostly used in event driven distributed systems"
    ),
    long_description = readme,
    long_description_content_type = "text/markdown",
    url = "https://github.com/bl17zar/pygrouper",
    packages = find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires = '>=3.6',
)
